from calendar import c


class SubParameter:
    pass


class Workflow:
    pass


class Element:
    pass


class TaskTemplate:
    pass


class TaskSchema:
    pass


class WorkflowTemplate:
    pass


class InputValue:
    pass


class Parameter:
    pass


class ParameterPath:
    pass


class Task:
    pass
